@@ -14,7 +14,7 @@ BEGIN {
 
 use Graph::AdjacencyMap qw(:flags :fields);
 
-our $VERSION = '0.9731';
+our $VERSION = '0.9733';
 
 require 5.006; # Weak references are absolutely required.
 
@@ -421,8 +421,15 @@ sub delete_vertex_by_id {
     &expect_non_unionfind;
     my ($g, $v, $id) = @_;
     return $g unless &has_vertex_by_id;
-    # TODO: what to about the edges at this vertex?
-    # If the multiness of this vertex goes to zero, delete the edges?
+    if ($g->[ _V ]->get_multi_ids( $v ) == 1) {
+      # only incarnation, zap edges
+      my @i = &_vertex_ids_multi;
+      pop @i; # the id
+      my $E = $g->[ _E ];
+      my @edges = $E->paths_from(@i);
+      push @edges, $E->paths_to(@i) if !&is_undirected;
+      $E->del_path( $_ ) for @edges;
+    }
     $g->[ _V ]->del_path_by_multi_id( $v, $id );
     $g->[ _G ]++;
     return $g;
@@ -984,6 +991,34 @@ sub rename_vertices {
     return $g;
 }
 
+sub filter_vertices {
+  my ($g, $code) = @_;
+  my @v = &_vertices05;
+  if (&is_multivertexed) {
+    for my $v (@v) {
+      $g->delete_vertex_by_id($v, $_) for
+        grep !$code->($g, $v, $_), $g->get_multivertex_ids($v);
+    }
+  } else {
+    $g->delete_vertices(grep !$code->($g, $_), @v);
+  }
+  $g;
+}
+
+sub filter_edges {
+  my ($g, $code) = @_;
+  my @e = &_edges05;
+  if (&is_multiedged) {
+    for my $e (@e) {
+      $g->delete_edge_by_id(@$e, $_) for
+        grep !$code->($g, @$e, $_), $g->get_multiedge_ids(@$e);
+    }
+  } else {
+    $g->delete_edges(map @$_, grep !$code->($g, @$_), @e);
+  }
+  $g;
+}
+
 sub as_hashes {
     my ($g) = @_;
     my (%v, %e, @e);
@@ -1112,6 +1147,32 @@ sub complete_graph {
     return $c;
 }
 
+sub max_cliques {
+    my ($g) = @_;
+    &expect_undirected;
+    $g->bron_kerbosch_pivot([], [$g->vertices], [], \ my @cliques);
+    return wantarray ? @cliques : \@cliques
+}
+
+sub bron_kerbosch_pivot {
+    my ($g, $r, $p, $x, $max_cliques) = @_;
+    if (! @$p && ! @$x && @$r) {
+        push @$max_cliques, [@$r];
+        return;
+    }
+    my $pivot = (@$p, @$x)[0];
+    for my $v (my @p = @$p) {
+        next if $g->has_edge($pivot, $v);
+        $g->bron_kerbosch_pivot(
+            [@$r, $v],
+            [grep { my $w = $_; grep $_ eq $w, @$p } $g->neighbours($v)],
+            [grep { my $w = $_; grep $_ eq $w, @$x } $g->neighbours($v)],
+            $max_cliques);
+        @$p = grep $_ ne $v, @$p;
+        push @$x, $v;
+    }
+}
+
 *complement = \&complement_graph;
 
 sub complement_graph {
@@ -1480,6 +1541,8 @@ sub random_graph {
     my %opt = _get_options( \@_ );
     __carp_confess "Graph::random_graph: argument 'vertices' missing or undef"
 	unless defined $opt{vertices};
+    __carp_confess "Graph::random_graph: both arguments 'edges' and 'edges_fill' specified"
+	if exists $opt{edges} && exists $opt{edges_fill};
     srand delete $opt{random_seed} if exists $opt{random_seed};
     my $random_edge = delete $opt{random_edge};
     my @V;
@@ -1493,26 +1556,20 @@ sub random_graph {
     delete $opt{vertices};
     my $V = @V;
     my $C = $V * ($V - 1) / 2;
-    my $E;
-    __carp_confess "Graph::random_graph: both arguments 'edges' and 'edges_fill' specified"
-	if exists $opt{edges} && exists $opt{edges_fill};
-    $E = exists $opt{edges_fill} ? $opt{edges_fill} * $C : $opt{edges};
-    delete $opt{edges};
-    delete $opt{edges_fill};
+    my $E = exists $opt{edges_fill} ? $opt{edges_fill} * $C : $opt{edges};
+    delete @opt{qw(edges edges_fill)};
     my $g = $class->new(%opt);
     $g->add_vertices(@V);
     return $g if $V < 2;
-    $C *= 2 if my $is_directed = $g->directed;
+    $C *= 2 if $g->directed;
     $E = $C / 2 unless defined $E;
     $E = int($E + 0.5);
     my $p = $E / $C;
     $random_edge = sub { $p } unless defined $random_edge;
-    # print "V = $V, E = $E, C = $C, p = $p\n";
     __carp_confess "Graph::random_graph: needs to be countedged or multiedged ($E > $C)"
 	if $p > 1.0 && !($g->countedged || $g->multiedged);
     # Shuffle the vertex lists so that the pairs at
     # the beginning of the lists are not more likely.
-    my (%v1_v2, @edges);
     my @V1 = _shuffle @V;
     my @V2 = _shuffle @V;
  LOOP:
@@ -1522,17 +1579,15 @@ sub random_graph {
 		next if $v1 eq $v2; # TODO: allow self-loops?
 		my $q = $random_edge->($g, $v1, $v2, $p);
 		if ($q && ($q == 1 || rand() <= $q) &&
-		    !exists $v1_v2{$v1}{$v2} &&
-		    ($is_directed ? 1 : !exists $v1_v2{$v2}{$v1})) {
-		    $v1_v2{$v1}{$v2} = undef;
-		    push @edges, [ $v1, $v2 ];
+		    !$g->has_edge($v1, $v2)) {
+		    $g->add_edge($v1, $v2);
 		    $E--;
 		    last LOOP unless $E;
 		}
 	    }
 	}
     }
-    $g->add_edges(@edges);
+    $g;
 }
 
 sub random_vertex {
@@ -1772,18 +1827,14 @@ sub _copy_edges {
   }
 }
 
-sub _undirected_copy_compute {
+sub undirected_copy {
+  &expect_directed;
   my $gc = $_[0]->new(undirected=>1);
   _copy_vertices($_[0], $gc);
   _copy_edges($_[0], $gc);
   $gc;
 }
 
-sub undirected_copy {
-  &expect_directed;
-  _check_cache($_[0], 'undirected_copy', [], \&_undirected_copy_compute);
-}
-
 *undirected_copy_graph = \&undirected_copy;
 
 sub undirected_copy_attributes {
@@ -1795,18 +1846,14 @@ sub undirected_copy_attributes {
   $gc;
 }
 
-sub _directed_copy_compute {
+sub directed_copy {
+  &expect_undirected;
   my $gc = $_[0]->new(undirected=>0);
   _copy_vertices($_[0], $gc);
   _copy_edges($_[0], $gc, 0, 1);
   $gc;
 }
 
-sub directed_copy {
-  &expect_undirected;
-  _check_cache($_[0], 'directed_copy', [], \&_directed_copy_compute);
-}
-
 *directed_copy_graph = \&directed_copy;
 
 sub directed_copy_attributes {
@@ -1874,11 +1921,10 @@ my %_cache_type =
     (
      'connectivity'        => ['_ccc'],
      'strong_connectivity' => ['_scc'],
+     'weak_connectivity_undirected_graph' => ['_wcug'],
      'biconnectivity'      => ['_bcc'],
      'SPT_Dijkstra'        => ['_spt_di', 'SPT_Dijkstra_root'],
      'SPT_Bellman_Ford'    => ['_spt_bf', 'SPT_Bellman_Ford_root'],
-     'undirected_copy'     => ['_undirected'],
-     'directed_copy'       => ['_directed'],
      'transitive_closure_matrix' => ['_tcm'],
     );
 
@@ -2024,33 +2070,40 @@ sub is_weakly_connected {
 
 *weakly_connected = \&is_weakly_connected;
 
+# because recreating undirected copy every time has different hash ordering
+# so weakly_connected_component_by_index etc would be unstable
+sub _weakly_connected_undir_graph {
+    _check_cache($_[0], 'weak_connectivity_undirected_graph', [],
+			   \&undirected_copy);
+}
+
 sub weakly_connected_components {
     &expect_directed;
-    splice @_, 0, 1, &undirected_copy;
+    splice @_, 0, 1, &_weakly_connected_undir_graph;
     goto &connected_components;
 }
 
 sub weakly_connected_component_by_vertex {
     &expect_directed;
-    splice @_, 0, 1, &undirected_copy;
+    splice @_, 0, 1, &_weakly_connected_undir_graph;
     goto &connected_component_by_vertex;
 }
 
 sub weakly_connected_component_by_index {
     &expect_directed;
-    splice @_, 0, 1, &undirected_copy;
+    splice @_, 0, 1, &_weakly_connected_undir_graph;
     goto &connected_component_by_index;
 }
 
 sub same_weakly_connected_components {
     &expect_directed;
-    splice @_, 0, 1, &undirected_copy;
+    splice @_, 0, 1, &_weakly_connected_undir_graph;
     goto &same_connected_components;
 }
 
 sub weakly_connected_graph {
     &expect_directed;
-    splice @_, 0, 1, &undirected_copy;
+    splice @_, 0, 1, &_weakly_connected_undir_graph;
     goto &connected_graph;
 }
 
